Package: SCArray
Type: Package
Title: Large-scale single-cell omics data manipulation with GDS files
Version: 1.19.0
Date: 2024-03-24
Depends: R (>= 3.5.0), gdsfmt (>= 1.36.0), methods, DelayedArray (>=
        0.31.5)
Imports: S4Vectors, utils, Matrix, SparseArray (>= 1.5.13),
        BiocParallel, DelayedMatrixStats, SummarizedExperiment,
        SingleCellExperiment, BiocSingular
Suggests: BiocGenerics, scater, scuttle, uwot, RUnit, knitr, markdown,
        rmarkdown, rhdf5, HDF5Array
Authors@R: c(person("Xiuwen", "Zheng", role=c("aut", "cre"),
        email="xiuwen.zheng@abbvie.com",
        comment=c(ORCID="0000-0002-1390-0708")))
Description: Provides large-scale single-cell omics data manipulation using
        Genomic Data Structure (GDS) files. It combines dense and sparse
        matrices stored in GDS files and the Bioconductor infrastructure
        framework (SingleCellExperiment and DelayedArray) to provide
        out-of-memory data storage and large-scale manipulation using the R
        programming language.
License: GPL-3
VignetteBuilder: knitr
ByteCompile: TRUE
URL: https://github.com/AbbVie-ComputationalGenomics/SCArray
biocViews: Infrastructure, DataRepresentation, DataImport, SingleCell,
        RNASeq
git_url: https://git.bioconductor.org/packages/SCArray
git_branch: devel
git_last_commit: d488bb9
git_last_commit_date: 2025-10-29
Repository: Bioconductor 3.23
Date/Publication: 2025-11-03
NeedsCompilation: yes
Packaged: 2025-11-04 00:40:58 UTC; biocbuild
Author: Xiuwen Zheng [aut, cre] (ORCID:
    <https://orcid.org/0000-0002-1390-0708>)
Maintainer: Xiuwen Zheng <xiuwen.zheng@abbvie.com>
